// AbstractAutomaton.hpp: this file is part of the REGAL project.
//
// REGAL : Random and Exhaustive Generators for Automata - Library
//
// Copyright (C) 2007 Julien DAVID.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// The complete GNU General Public Licence Notice can be found as the
// `COPYING' file in the root directory.
//
//
#ifndef ABSTRACTAUTOMATON
#define ABSTRACTAUTOMATON


#include "GasteXConverter.hpp"
#include "DotConverter.hpp"
#include "Converter.hpp"
#include "VerboseOption.hpp"
#include "Alphabet.hpp"

#include<map>

namespace regal{

  struct InvalidStateException : public std::exception {};
  
  template<typename StateLabel_t,typename Sigma>
  class AbstractAutomaton{
  private:
    int stateNumber; /*!< actual number of states in the automaton */
    std::map<int,StateLabel_t> intTostateMap; /*!< Used to make sure every state is associated to an integer */
    std::map<StateLabel_t,int> stateTointMap; /*!< Used to get the numerical value associated to a state */
    
    /**
       Adds a State in the Automaton 
       @Return the created state
    */
    virtual StateLabel_t createState()=0;
    
  protected:
    Alphabet<Sigma> alphabet;
    
    
    
  public:
    /**
       Creates an Abstract Automaton
    */
    AbstractAutomaton(const Alphabet<Sigma> & a):alphabet(a){
      stateNumber=0;
      verbose("Creation of an Automaton");
    }
    
    /**
       Destroy the current Abstract Automaton
    */
    virtual ~AbstractAutomaton(){
      verbose("Destruction of an Automaton");
    }
    
    virtual Alphabet<Sigma> getAlphabet()const{return alphabet;}
    
    
    virtual int getSize()=0;
    
    virtual int getAlphabetSize(){return alphabet.getSize();}
    
    /**
       Adds a State in the Automaton and associate it to a numerical value
       @Return the created state
    */
    virtual StateLabel_t addState(){
      intTostateMap[stateNumber++]=createState();
      stateTointMap[intTostateMap[stateNumber-1]]=stateNumber-1;
      return intTostateMap[stateNumber-1];
    }
    
    int getIntegerValue(const StateLabel_t & s){
      return stateTointMap[s];
    }
    
    StateLabel_t getRealValue(const int & ind){
      return intTostateMap[ind];
    }
    
    /**
       Adds a Transition in the Automaton
       
       @param start is the label of the state where the new transition starts
       @param end is the label of the state where the new transition ends
       @param value is the value of the transition
    */
    virtual void addTransition(const StateLabel_t & start,const StateLabel_t & end,const Sigma & value)=0;
    
    /**
       Tells if a state is final
       
       @param sl is the label of the state
       @Return True if the state is final, False otherwise
    */
    virtual bool isFinal(const StateLabel_t & sl)=0;
    
    /**
       Tells if a state is initial
       
       @param sl is the label of the state
       @Return True if the state is initial, False otherwise
    */
    virtual bool isInitial(const StateLabel_t & sl)=0;
    
    /**
       Set a state as final
       @param sl is the label of the state
       @param value: if true, s1 is final.
    */
    virtual void setStateAsFinal(const StateLabel_t & sl,const bool & value)=0;
    /**
       Set a state as initial
       @param sl is the label of the state
       @param value, if true, s1 is initial. 
    */
    virtual void setStateAsInitial(const StateLabel_t & sl,const bool & value)=0;
  };
  
}

#endif
