// DFAAutomaton.hpp: this file is part of the REGAL project.
//
// REGAL : Random and Exhaustive Generators for Automata - Library
//
// Copyright (C) 2007 Julien DAVID.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// The complete GNU General Public Licence Notice can be found as the
// `COPYING' file in the root directory.
//
//
#ifndef DFAAUTOMATON
#define DFAAUTOMATON

#include <list>
#include <iterator>
#include "AbstractAutomaton.hpp"
#include "VerboseOption.hpp"


using namespace std;

namespace regal{

  struct AutomatonLimitSizeExcedeedException : public std::exception {};
  
  template<typename StateLabel_t,typename Sigma>
  class DFAAutomaton: public AbstractAutomaton<StateLabel_t, Sigma> {
  private:
    int maxSize; /*!< max number of states in the automaton */
    int number; /*!< actual number of states in the automaton */
    int languageSize; /*!< alphabet size */
    int ** data; /*!< matrix of transitions */
    bool * finalStates; /*!< Tabular that indicates if a state is final */
    bool * initialStates; /*!< Tabular that indicates if a state is initial */
  
    
    /**
     Export the DFA into a grapgical format
     @param cc is the converter (can be a DoTConverter, GasteXConverter)...
    */
    void exportFormat(Converter * cc){
      char * name=(char*)calloc(sizeof(char),32);
      char * name2=(char*)calloc(sizeof(char),32);
      char * word=(char*)calloc(sizeof(char),32);
      int c,w,c2;
      std::list<Sigma> listT=this->alphabet.getList();
      typename std::list<Sigma>::iterator l;
      cc->beginAutomaton(165,(maxSize/5)*33);
      
      for(c=0;c<getSize();c++){
	sprintf(name,"%d",c);
	cc->draw_node(name,(c%5)*20,-(c/5)*33,isInitial(c),isFinal(c));
      }
    
      for(c=0;c<getSize();c++){
	//for(w=0;w<languageSize;w++){
	for(l=listT.begin();l!=listT.end();l++){  
	  w=this->alphabet.getNumericalValue(*l);
	  c2=data[c][w];
	  if(c2!=-1){
	    sprintf(name,"%d",c);
	    sprintf(name2,"%d",c2);
	    sprintf(word,"%c",*l);
	    cc->draw_edge(name,name2,word);
	  }
	}
      }
      free(name);
      free(name2);
      free(word);
      cc->endAutomaton();
    }
    
  public:
    /**
       This function is useless, as states are created
       in the Contructor.
       @Return the state added by the automaton
    */
    int createState(){
      if(number<maxSize){
	//StateLabel_t res=(StateLabel_t)number++;
	//return res;
	return number++;
      }
      cout<<"This Automaton has a fixed number of states"<<endl;
      throw new AutomatonLimitSizeExcedeedException();
    }
    
    
    /**
       Returns the actual size of the automaton
       @Return the actual size of the automaton
    */
    int getSize(){return number;}
    
    /**
       Returns q, with p-->w-->q
       @param start is the starting state
       @param word is the label on the transition coming out from start
       @Return q, with p-->w-->q
    */
    StateLabel_t getArrivalState(const StateLabel_t & start,const Sigma & word){
      return getRealValue(data[getIntegerValue(start)][this->alphabet.getNumericalValue(word)]);
    }
    
    /**
       Returns q, with p-->w-->q
       @param start is the starting state
       @param word is the label on the transition coming out from start
       @Return q, with p-->w-->q
    */
    int getArrivalNumericalState(const int & start,const Sigma & word){
      return data[start][this->alphabet.getNumericalValue(word)];
    }
    
    /**
       Adds a transtition between start and end.
       @param start is the state where the transition comes from.
       @param end is the state pointed by the transition
       @param value is the label associated to the transition.
    */
    void addTransition(const StateLabel_t& start,const StateLabel_t& end,
		       const Sigma & value){
      if(start<maxSize&&end<maxSize&&this->alphabet.getNumericalValue(value)<languageSize&&end>=0)
	data[start][this->alphabet.getNumericalValue(value)]=end;
      else{
	cout<<"Automaton Max Size : "<<maxSize<<", tried values : "<<start<<" "<<end<<" "<<value<<endl;
	throw new InvalidStateException();
      }
    }
    
    /**
       Checks if a state is Final
       @param s1 is the state we want to verify
       @Return true if s1 is final, false otherwise
    */
    bool isFinal(const StateLabel_t & sl){
      if(sl<maxSize&&finalStates[sl]==true)
	return true;
      return false;
    }

    /**
       Checks if a state is initial
       @param s1 is the state we want to verify
       @Return true if s1 is initial, false otherwise
    */
    bool isInitial(const StateLabel_t & sl){
      if(sl<maxSize&&initialStates[sl]==true)
	return true;
      return false;
    }
    
    /**
       Set a state as final
       @param s1 is the state to parameter
    */
    void setStateAsFinal(const StateLabel_t & sl,const bool & value){
      if(sl<maxSize)
	finalStates[sl]=value;
      else
	throw new InvalidStateException();
    }
    
    /**
       Set a state as initial
       @param s1 is the state to parameter
    */
    void setStateAsInitial(const StateLabel_t & sl,const bool & value){
      if(sl<maxSize){
	initialStates[sl]=value;
      }
      else
	throw new InvalidStateException();
    }
    
    
    /**
       Translate a DFA into gasteX file format, in order 
       to have a graphical view of the object.
       @param gc is the object which generate gasteX code.
    */
    void toGasteX(GasteXConverter * gc){
      exportFormat(gc);
    }
    
    
    /**
       Convert the current automaton into a DOT file 
       @param F is the file where the DOT format automaton will be saved
    */
    void toDOT(DotConverter * dc){
      exportFormat(dc);
    }
    
    
    friend ostream& operator<<(ostream& o,const DFAAutomaton<int,char> & a);
    
    /**
       Creates a DFAAutomaton
    */
    DFAAutomaton(const int& sSize,const Alphabet<Sigma> & a):AbstractAutomaton<StateLabel_t,Sigma>::AbstractAutomaton(a){
      int i,j;
      maxSize=sSize;
      languageSize=a.getSize();
      number=0;
      
      data= new int*[maxSize];
      finalStates= new bool[maxSize];
      initialStates= new bool[maxSize];
      
      for(i=0;i<maxSize;i++){
	data[i]=new int[languageSize];
	for(j=0;j<languageSize;j++)
	  data[i][j]=-1;
	finalStates[i]=false;
	initialStates[i]=false;
      }
      verbose("Creation of an DFAAutomaton");
      
    }
    /**
       Destroy the current DFAAutomaton
       @param sSize is the number of states
       @param lSize if the size of the alphabet
    */
    virtual ~DFAAutomaton(){
      int i;
      for(i=0;i<maxSize;i++)
	delete [] data[i];
      delete [] data;
      delete [] finalStates;
      delete [] initialStates;
      verbose("Destruction of an DFAAutomaton");
    }
  };
  


  /**
     Print a DFA automaton
     @param o is the stream in which we add the DFA's description
     @param a is the DFA we want to print
     @Return the modified stream 
  */
  ostream& operator<<(ostream& o,const DFAAutomaton<int,char> & a){
    int i,j;
    o<<endl<<"----------"<<endl;
    o<<"Automaton"<<endl;
    o<<"MaxSize = "<<a.maxSize<<endl;
    o<<"Language Size = "<<a.languageSize<<endl<<endl;
    o<<"   ";
    std::list<char> listT=a.getAlphabet().getList();
    for(std::list<char>::iterator l=listT.begin();l!=listT.end();l++)
      o<<"  "<<*l;
    o<<"    Initial    Final";
    o<<endl<<endl;
    for(i=0;i<a.number;i++){
      o<<i<<"  ";
      for(j=0;j<a.languageSize;j++)
	if(a.data[i][j]==-1)
	  o<<"  "<<(char)a.data[i][j];
	else
	  o<<"  "<<a.data[i][j];
      o<<"    "<<a.initialStates[i]<<"          "<<a.finalStates[i];
      o<<endl;
    }
    o<<endl<<"----------"<<endl;
    return o;
  }

}

#endif
