// MooreAlgorithm.hpp: this file is part of the REGAL project.
//
// REGAL : Random and Exhaustive Generators for Automata - Library
//
// Copyright (C) 2007 Julien DAVID.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// The complete GNU General Public Licence Notice can be found as the
// `COPYING' file in the root directory.
//
//
#ifndef MOOREALGORITHM
#define MOOREALGORITHM

#include "MinimizingAlgorithm.hpp"
#include "Alphabet.hpp"
#include "toolbox/GenericFunction.hpp"
#include "toolbox/LexicographicalSort.hpp"


namespace regal{
  
  template<typename StateLabel_t,typename Sigma,class Automaton_t=AbstractAutomaton<StateLabel_t,Sigma> >
  class MooreAlgorithm: public MinimizingAlgorithm<StateLabel_t,Sigma,Automaton_t>{
  private:
    int *partitionTmp;
    int ** partitionWord; 
    int complexity; /*!< Used to measure Moore's complexity */
    int * tmpClassNum[2]; /*!< Used to measure how fast the classes are divided */
    bool tmpClassNumFlag;
    int aSize; /*!< Size of the alphabet */
    List_t ** lexicographical;
    List_t * lexicographicalRes;
    List_t * lexicographicalTMP;

    
    
    void freeMemory(){
      if(this->partitionRes!=NULL)delete [] this->partitionRes;
      if(partitionTmp!=NULL)delete [] partitionTmp;
      if(partitionWord!=NULL){
	for(int i=0;i<this->size;i++){
	  delete [] partitionWord[i];
	  destroyList(lexicographical[i]);
	}
	free(lexicographical);
	destroyList(lexicographicalRes);
	destroyList(lexicographicalTMP);
	delete [] partitionWord;
      }
    }
    
    void initPartition(const int & s,const int & as){
      if(s!=this->size){
	freeMemory();
	this->size=s;
	aSize=as;
	this->partitionRes=new int[this->size];
	partitionTmp=new int[this->size];
	partitionWord=new int*[this->size];
	lexicographical=(List_t **)malloc(sizeof(List_t *)*s);
	for(int i=0;i<this->size;i++){
	  partitionWord[i]=new int[as];
	  lexicographical[i]=createList();
	}
	lexicographicalTMP=createList();
	lexicographicalRes=createList();
      }
    }
    
    void lexicographicalSort(int prof){
      doubleChainedList_t * tmp;
      while(lexicographicalRes->first!=NULL){
	tmp=extractFirstList(&lexicographicalRes);
	if(prof==0)
	  insertNodeLastList(&lexicographical[this->partitionRes[tmp->val]],tmp);
	else
	  insertNodeLastList(&lexicographical[this->partitionWord[tmp->val][prof-1]],tmp);
      }
      for(int i=0;i<this->maxClass;i++)
	concatList(&lexicographicalRes,&lexicographical[i]);    
    }
    
    void sortClasses(){
      doubleChainedList_t * tmp,*tmp2;
      if(tmpClassNumFlag==true)
	memset(tmpClassNum[1],0,this->size*sizeof(int));
      for(int i=aSize;i>=0;i--)
	lexicographicalSort(i);
      this->maxClass=1;
      
      for(tmp=lexicographicalRes->first,tmp2=NULL;tmp!=NULL;tmp2=tmp,tmp=tmp->next){
	if(tmp2!=NULL&&!sameClass(aSize,tmp->val,tmp2->val))
	  this->maxClass++;
	partitionTmp[tmp->val]=this->maxClass-1;
	if(tmpClassNumFlag==true)
	  tmpClassNum[1][this->maxClass]++;
      }
    }
    
    bool mooreFirstCutting(Automaton_t * a){
      int i;
      bool res=true;
      for(i=0;i<this->size;i++){
	if(a->isFinal(a->getRealValue(i))){
	  this->partitionRes[i]=0;
	  insertLastList(&lexicographicalRes,i);
	}
	else{
	  this->partitionRes[i]=1;
	  insertLastList(&lexicographicalTMP,i);
	}
      }
      if(lexicographicalRes->first==0||lexicographicalTMP->size==0)
	res=false;
      concatList(&lexicographicalRes,&lexicographicalTMP);
      this->maxClass=2;
      return res;
    }
    
    bool sameClass(const int & alphabetSize,const int & state1,const int & state2){
      int i;
      if(this->partitionRes[state1]!=this->partitionRes[state2])return false;
      for(i=0;i<alphabetSize;i++)
	if(partitionWord[state1][i]!=partitionWord[state2][i])
	  return false;
      return true;
    }
    
    bool isPartitionMinimal(){
      int i;
      for(i=0;i<this->size;i++)
	if(this->partitionRes[i]!=partitionTmp[i])
	  return false;
      return true;
    }
    
  public:
    
    /**
       Minimize an automaton with Moore algorithm and returns a partition of the states.
       @param a is the automaton to minimize
       @Return partition of the states, in which i is a states of a and p[i]
       is a state of Moore(a)
    */
    int * minimizeToPartition(Automaton_t * a,const int & Mit=-1){
      int i;
      Alphabet<Sigma> al=a->getAlphabet();
      std::list<Sigma> listT=al.getList();
      initPartition(a->getSize(),al.getSize());
      complexity++;
      if(!mooreFirstCutting(a)){
	emptyList(lexicographicalRes);
	return this->partitionRes;
      }
      /*std::cout<<"-----------------------------------"<<std::endl;
      for(i=0;i<this->size;i++)
	std::cout<<this->partitionRes[i]<<" ";
      std::cout<<std::endl;
      */
      while(1){
	complexity++;
	for(i=0;i<this->size;i++){
	  for(std::list<char>::iterator l=listT.begin();l!=listT.end();l++)
	    partitionWord[i][al.getNumericalValue(*l)]=this->partitionRes[a->getArrivalState((StateLabel_t)i,*l)];
	}
	sortClasses();
	if(tmpClassNumFlag==true)
	  tmpClassNum[0][complexity-1]=this->maxClass;
	
	swapT(this->partitionRes,partitionTmp);
	/*for(i=0;i<this->size;i++)
	  std::cout<<this->partitionRes[i]<<" ";
	std::cout<<std::endl;
	*/
	if(this->maxClass==a->getSize()||isPartitionMinimal()||((complexity-1)==Mit)){
	  emptyList(lexicographicalRes);
	  return this->partitionRes;
	}
      }
      
    }

    /**
       Minimize an automaton with Moore algorithm and returns the class number at each iteration.
       @param a is the automaton to minimize
       @Return a tabular containing the class number at each iteration (first value is the size of the tabular)
       and the number of states in each classes.
    */
    int ** minimizeSteps(Automaton_t * a,const int & precision=-1){
      if(tmpClassNum[0]!=NULL){
	delete [] tmpClassNum[0];
	delete [] tmpClassNum[1];
      }
      tmpClassNum[0]=new int[a->getSize()];
      tmpClassNum[1]=new int[a->getSize()];
      complexity=0;
      tmpClassNumFlag=true;
      minimizeToPartition(a,precision);
      tmpClassNumFlag=false;
      tmpClassNum[0][0]=complexity;
      tmpClassNum[1][0]=this->maxClass;
      return tmpClassNum;
    }

       
    /**
       Minimize an automaton with Moore algorithm and only returns the size of the result.
       @param a is the automaton to minimize
       @Return the size of the result.
    */
    int minimizeSize(Automaton_t * a){
      minimizeToPartition(a);
      return this->maxClass;
    }

    /**
       Minimize an automaton with Moore algorithm 
       @param a is the automaton to minimize
       @Return the minimized automaton
    */
    Automaton_t * minimize(Automaton_t * a){
      int i,j;
      Alphabet<Sigma> al=a->getAlphabet();
      std::list<Sigma> listT=al.getList();
      minimizeToPartition(a);
      Automaton_t * res=new Automaton_t(this->maxClass,al);
      //Transform the partition into an Automaton
      //for(i=0;i<a->getSize();i++)
      //std::cout<<this->partitionRes[i]<<" ";
      //std::cout<<std::endl;
      for(i=0;i<this->maxClass;i++)
	res->addState();
      for(i=0,j=0;i<this->size;i++){
	j=this->partitionRes[i];
	if(a->isInitial(i))
	  res->setStateAsInitial(j,true);
	if(a->isFinal(i))
	  res->setStateAsFinal(j,true);
	for(std::list<char>::iterator l=listT.begin();l!=listT.end();l++)
	  res->addTransition(this->partitionRes[i],partitionWord[i][al.getNumericalValue(*l)],*l);
	
      
      }
      return res;
    }
    
    
    /**
       Create a instance of MooreAlgorithm, an optimized class for minimizing automaton
     */
    MooreAlgorithm(){
      this->size=0;
      this->partitionRes=NULL;
      partitionTmp=NULL;
      partitionWord=NULL;
      tmpClassNum[0]=NULL;
      tmpClassNum[0]=NULL;
      tmpClassNumFlag=false;
    }
    
    /**
       Destroy a MooreAlgorithm instance
     */
    ~MooreAlgorithm(){
      freeMemory();
      if(tmpClassNum[0]!=NULL){delete [] tmpClassNum[0];delete [] tmpClassNum[1];}
    }
    
    /**
       Returns the complexity of the minimization on a
       @param a is the automaton to minimize
       @Return the complexity of the minimization
    */
    int minimizeComplexity(Automaton_t * a){
      complexity=0;
      minimizeToPartition(a);
      return complexity;
    }
    unsigned int * minimizeMultipleComplexity(Automaton_t * a){
      return NULL;
    }
    
  };

}



#endif
