// RandomDFAGenerator.hpp: this file is part of the REGAL project.
//
// REGAL : Random and Exhaustive Generators for Automata - Library
//
// Copyright (C) 2007 Julien DAVID.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// The complete GNU General Public Licence Notice can be found as the
// `COPYING' file in the root directory.
//
//
#ifndef RANDOMDFAGENERATOR
#define RANDOMDFAGENERATOR

#include "AbstractRandomGenerator.hpp"
#include "AbstractDFAGenerator.hpp"
#include "BoltzmannSampler.hpp"
#include "toolbox/GenericFunction.hpp"
#include "VerboseOption.hpp"
#include "MooreAlgorithm.hpp"
#include "StronglyConnected.hpp"
#include "CoAccessible.hpp"


namespace regal{
  
  template<typename StateLabel_t,typename Sigma, class Automaton_t=AbstractAutomaton<StateLabel_t,Sigma> >
  class RandomDFAGenerator : 
    public AbstractRandomGenerator<Automaton_t>, AbstractDFAGenerator<StateLabel_t,Sigma,Automaton_t>{
  private:
    BoltzmannSampler * bs; /*!< Gives a partition that will be converted into an automaton */
    int *partBoltzmann; /*!< Partition given by bs, */
    int * partMin;
    int ** partFinal; /*!< Subsets, whose size are given by partBoltzmann */
    int * catS; /*!< CatalanSuit - Word of Dyck : used to create a trie */
    int * comS; /*!< CompletionSuit : used to complete the trie into a DFA */ 
    int * partTMP; /*!< used during the conversion from partFinal to catS and comS */
    int rejectCounters[2];
    double probability;

    /**
       Sort sub-sets of a partition according to their smallest element.
    */
    void sortPartition(){
      int min,i,j,tmp,*tmpL;
      int size=this->result->getSize();
      for(i=1;i<size;i++){
	min=partMin[i];
	for(j=i;j>0;j--)
	  if(min<partMin[j-1]){
	    tmpL=partFinal[j-1];
	    partFinal[j-1]=partFinal[j];
	    partFinal[j]=tmpL;
	    tmp=partBoltzmann[j-1];
	    partBoltzmann[j-1]=partBoltzmann[j];
	    partBoltzmann[j]=tmp;
	    tmp=partMin[j-1];
	    partMin[j-1]=partMin[j];
	    partMin[j]=tmp;
	  }
      }
    }
    
    /**
       Prints a partition, which is a 2-dimensional tabular
    */
    void printPartition(){
      int i,j;
      int size=this->result->getSize();
      for(i=0;i<size;i++){
	std::cout<<"{";
	for(j=0;j<partBoltzmann[i];j++)
	  std::cout<<partFinal[i][j]<<",";
	std::cout<<"}";
	
      }
      std::cout<<std::endl;
    }
    
    /**
       Generate a random permutation
       @param max is the size of the permutation
       @Return the generated permutation
     */
    int * randomPermutation(const int & max){
      int i,pos;
      int * l=new int[max];
      for(i=0;i<max;i++){
	pos=rand()%(i+1);
	if(pos!=i)
	  l[i]=l[pos];
	l[pos]=i+1;
      }
      return l;
    }
    
    /**
       Part of the Bassino-Nicaud algorithm:
       Generate a random partition, using a Bolztmann Sampler.
    */
    void randomPartition(){
      int max=this->result->getSize()*this->result->getAlphabetSize(),i,j,counter,pos;
      int * l=randomPermutation(max);
      memset(partMin,0,this->result->getSize()*sizeof(int));
      
      partBoltzmann=bs->sample();
      
      for(pos=0,i=0,counter=0;counter<max;i++){
	partFinal[i]=new int[partBoltzmann[i]];
	for(j=0;j<partBoltzmann[i];j++,pos++){
	  partFinal[i][j]=l[pos];
	  if(partMin[i]==0)partMin[i]=partFinal[i][j];
	  else if(partFinal[i][j]<partMin[i])partMin[i]=partFinal[i][j];
	  max--;
	}
      }
      sortPartition();
      delete [] l;
    }
    
    
    /**
       Part of the Bassino-Nicaud algorithm: translate a partition
       into 2 Suits : catS which is a word of Dyck and comS
       which complete catS in the automaton's contruction.
    */
    void partitionToSuits(){
      int i,j,max=0,pos=0;
      int size=this->result->getSize();
      for(i=0;i<size;i++)
	for(j=0;j<partBoltzmann[i];j++)
	  partTMP[partFinal[i][j]-1]=i+1;
      
      for(i=0;i<(size*this->result->getAlphabetSize());i++){
	if(partTMP[i]>max)max=partTMP[i];
	else{
	  comS[pos]=partTMP[i];
	  catS[pos]=max;
	  pos++;
	}
      }
      //Adds the last transition without boltzmann to ensure equiprobability that has been
      //proved for automatonSize*alphabetSize
      catS[pos]=max;
      comS[pos]=rand()%max+1;
    }
    
    /**
       Free the memory space used by partFinal
    */
    void deletePartition(){
      int i;
      for(i=0;i<this->result->getSize();i++)
	delete [] partFinal[i];  
    }
    
    /**
       Randomdly assign the final states on the generated automaton
    */
    void randomFinalization(){
      int i;
      StateLabel_t state;
      for(i=0;i<this->result->getSize();i++){
	state=this->result->getRealValue(i);
	
	if(((rand()%RAND_MAX)/(double)RAND_MAX)<probability)
	  this->result->setStateAsFinal(state,true);
	else
	  this->result->setStateAsFinal(state,false);
      }
    }
    
    
    /**
       Randomdly design a state to be the only final state
    */
    void onlyOneFinalState(){
      int i;
      StateLabel_t state;
      for(i=0;i<this->result->getSize();i++){
	state=this->result->getRealValue(i);
	this->result->setStateAsFinal(state,false);
      }
      this->result->setStateAsFinal(this->result->getRealValue(rand()%this->result->getSize()),true);
    }
    

    /**
       Generate randomly a DFA, without finalizing it.
    */
    void randomUnfinalizedDFA(){
      rejectCounters[1]=0;
      do{
	randomPartition();
	partitionToSuits();
	deletePartition();
	rejectCounters[1]++;
      }
      while(!this->isACatalanSuit(catS));
#ifdef __REJECTFLAG
      std::cout<<"Nombre de rejet : "<<rejectCounters[1]<<std::endl;
#endif 
      this->suitToDFA(catS,comS);
    }

    bool isMinimal(Automaton_t * a){
      int i,max=0,size=a->getSize();
      MooreAlgorithm<int,char,Automaton_t > ma;
      int * partition=ma.minimizeToPartition(a);
      for(i=0;i<size;i++)
	if(partition[i]>max)
	  max=partition[i];
      if(max==(size-1))
	return true;
      return false;
    }

    
  public:
    
    /**
       Returns a randomly generated automaton
       @Return a randomly generated automaton
    */
    Automaton_t * random(){
      randomUnfinalizedDFA();
      randomFinalization();
      return this->result;
    }
    
    /**
       Returns a randomly generated minimal automaton
       @Return a randomly generated minimal automaton
    */
    Automaton_t * randomMinimal(){
      do{
	random();
      }while(!isMinimal(this->result));
      return this->result;
    }

    /**
       Returns a randomly generated strongly connected automaton
       @Return a randomly generated strongly connected automaton
    */
    Automaton_t * randomStronglyConnected(){
      do{
	random();
      }while(!StronglyConnectedTester<StateLabel_t,Sigma,Automaton_t>::testConnectability(this->result));
      return this->result;
    }
    
    /**
       Returns a randomly generated strongly connected automaton
       @Return a randomly generated strongly connected automaton
    */
    Automaton_t * randomCoAccessible(){
      do{
	random();
      }while(!CoAccessibleTester<StateLabel_t,Sigma,Automaton_t>::testCoAccessibility(this->result));
      return this->result;
    }

    /**
       Returns a randomly generated automaton with only one final state
       @Return a randomly generated automaton with only one final state
    */
    Automaton_t * randomOneFinalState(){
      randomUnfinalizedDFA();
      onlyOneFinalState();
      return this->result;
    }

    /**
       Returns the number of reject used to generate the previous Automaton
       @Return the number of Bolztmann rejects and Dyck rejects.
     */
    int * getRejectCounters(){
      rejectCounters[0]=bs->getRejectCounter();
      return rejectCounters;
    }
    
    /**
       Create a random DFA Generator
       @param autSize is the number of states in the generated automatons
       @param alphaSize is the size of the alphabet in the generated automatons
       @param prob is the probability for a state to be final
    */
    RandomDFAGenerator(const int & autSize,const Alphabet<Sigma> & alpha,const double & prob=0.5):
      AbstractDFAGenerator<StateLabel_t,Sigma,Automaton_t>::AbstractDFAGenerator(autSize,alpha){
      bs=new BoltzmannSampler(autSize,alpha.getSize());
      catS=new int[autSize*(alpha.getSize()-1)+1];
      comS=new int[autSize*(alpha.getSize()-1)+1];
      partTMP=new int[autSize*alpha.getSize()];
      partMin=new int[autSize];
      partFinal=new int*[autSize];
      probability=prob;
      
      verbose("Creation of the RandomDFAGenerator");
    }
    
    
    /**
       Destroy the RandomDFAGenerator          
    */
    ~RandomDFAGenerator(){
      delete bs;
      delete [] catS;
      delete [] comS;
      delete [] partTMP;
      delete [] partFinal;
      delete [] partMin;
      verbose("Destruction of the RandomDFAGenerator");
      
    }
    
  };

}

#endif
