//    This file is part of RoadColoringImpl.
//
//    Copyright (C) 2008 Florian SIKORA sikora@univ-mlv.fr
//    http://igm.univ-mlv.fr/~fsikora/index.php?p=r
//
//    RoadColoringImpl is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    RoadColoringImpl is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    The complete GNU General Public Licence Notice can be found as the
//    `COPYING' file in the root directory.


package fr.umlv.roadcoloring.graph;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.LinkedList;

/**
 * 
 * Compute the pair automata in order to find a	synchronize word
 * 
 * @author fsikora
 *
 */
public class Pairs {
	private final ArrayList<PairsState> states = new ArrayList<PairsState>();
	private final Graph g;
	
	private final static char RED = 'R';
	private final static char BLUE = 'B';
	
	private String syncWord = "";
	
	public Pairs(Graph g) {
		this.g = g;
	}
	
	/**
	 * Compute the pair automata in each step and return the synchronize word for the graph
	 * 
	 * 
	 * @return the synch word
	 */
	public String getSynchronizedWord() {
		//start state
		int a = 0;
		
		for (int i=1 ; i<g.getSize() ; ++i) {
			//synchronize a with i + current syncWord
			int b = g.browse(i, syncWord);
			
			if(a != b) {
				//construct pair automata : check a singleton who significate a syncWord, between a and b
				states.clear();
				PairsState set = new PairsState();
				set.addNode(a);
				set.addNode(b);
				states.add(set);
				//check the singleton and get the final state
				int res = getSingletonFromSet(set);
				
				//next a is the state
				a = res;
			}
		}
		return syncWord;
	}
	
	/**
	 * Return a new state iff it's don't exist before
	 * 
	 * @param prec the start state
	 * @param image image by color
	 * @param color color used to lead prec to image
	 * @return the new state (null if already exist)
	 */
	private PairsState addState(PairsState prec, HashSet<Integer> image, char color) {
		//check if there is a state who correspond to this image
		for (PairsState pairsState : states) {
			if (pairsState.equals(image)) {
				return null;
			}
		}
		//no existed found, create new state to trait
		PairsState pairsState = new PairsState(image, prec.getWord().concat(String.valueOf(color)));
		states.add(pairsState);
		
		return pairsState;
	}
	
	
	/**
	 * Perform the pair automata, update the syncWord, and return the state end
	 * @param set start set
	 * @return end node
	 */
	private int getSingletonFromSet(PairsState set) {
		//will browse by blue and by red
		HashSet<Integer> imageRed = new HashSet<Integer>();
		HashSet<Integer> imageBlue = new HashSet<Integer>();
		int outBlue = 0;
		int outRed = 0;
		String wordRed = "";
		String wordBlue = "";
		
		LinkedList<PairsState> toTrait = new LinkedList<PairsState>();
		toTrait.add(set);
		
		//perform and add to traitment list while we don't find a singleton state
		do {
			PairsState current = toTrait.remove();
			imageRed = new HashSet<Integer>();
			imageBlue = new HashSet<Integer>();

			//construct the images
			for (int node : current.getNodes()) {
				outRed = g.getRed(node);
				imageRed.add(outRed);
				outBlue = g.getBlue(node);
				imageBlue.add(outBlue);
			}
			
			//check if the image already exist. If not, create a new state
			PairsState next = addState(current, imageBlue, BLUE);
			if(next != null) {	//add in the traitment list and update the word
				toTrait.add(next);
				wordBlue = next.getWord();
			}
			next = addState(current, imageRed, RED);
			if(next != null) {
				toTrait.add(next);
				wordRed = next.getWord();
			}			
		} while (imageRed.size() > 1 && imageBlue.size() > 1 && toTrait.size() > 0);	//singleton founded
		
		//sucess by blue
		if (imageBlue.size() == 1) {
			syncWord = syncWord.concat(wordBlue);
			return outBlue;
		}
		//sucess by red
		else if (imageRed.size() == 1) {
			//System.out.println("ha " + wordRed);
			syncWord = syncWord.concat(wordRed);
			return outRed;
		}
		else {	//error, traitment list empty and no singleton founded
			throw new AssertionError("Singleton not found !");
		}
	}
}
